/*
   See pages 176-179, Figure 6.16, from "Principles of Parallel Programming".

   This version measures deltas but uses iterations for stopping.
*/
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "2D_SOR.h"
#include "initial_data_1.h"

#define MAXTHREADS 16      /* maximum number of threads */
void* thread_main(void*);  /* thread function prototype */
void barrier();            /* "home made" barrier prototype */

pthread_mutex_t update_lock;
pthread_mutex_t barrier_lock; /* mutex for the barrier CV */
pthread_cond_t all_here;      /* condition variable for barrier */
int count = 0;                /* counter for barrier */
int done = 0;                 /* used to terminate main do-loop */

/*
  Command line args: matrix size, number of threads, threashold
*/
int main(int argc, char*argv[])
{
   /* thread ids and attributes */
   pthread_t tid[MAXTHREADS];
   pthread_attr_t attr;
   int i, err;

   /* set global thread attributes */
   pthread_attr_init(&attr);
   pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

   /* initialize mutex and condition variables */
   pthread_mutex_init(&update_lock, NULL);
   pthread_mutex_init(&barrier_lock, NULL);
   pthread_cond_init(&all_here, NULL);

   /* read command line arguments */
   if (argc != 4)
   {
      printf("usage: %s n t iterations\n", argv[0]);
      return 0;
   }
   n = atoi(argv[1]);          /* number of rows (and columns) */
   t = atoi(argv[2]);          /* number of threads */
   iterations = atof(argv[3]); /* determines when to stop */
   rowsPerThread = (n-2) / t;  /* be sure t divides n-2 (or data will be lost) */
   if (n % t != 0)
   {
      printf("usage: %s n t iterations\n", argv[0]);
      printf("       Where t divides n.\n");
      return 0;
   }

   InitializeData(&new, &val, n);

   for(i = 0; i < t; i++)
   {
      err = pthread_create(&(tid[i]), &attr, thread_main, (void*)i);
      if (err != 0)
      {
         printf("pthread_create: error number %d\n", err);
         exit(1);
      }
   }
   for(i = 0; i < t; i++)
   {
      pthread_join(tid[i], NULL);
   }

   //printf("maximum difference: %e\n", delta);

   WriteData(); /* write the data to a file so that it can be visualized */
   printf("total iterations = %d\n", iterates);
   printf("final delta = %g\n", delta);

   return 0;
}/* main() */


void* thread_main(void* arg)
{
   int id = (int)arg;        /*** compiler warning with 64 bit compilers ***/
   double average, myDelta;
   double **myVal, **myNew;
   double **temp;
   int i, j;
   int start;

   /* determine first row that this thread owns */
   start = (id * rowsPerThread) + 1;
   myVal = val;
   myNew = new;

   do
   {
      myDelta = 0.0;

      /* update each point */
      for(i = start; i < start+rowsPerThread; i++)
      {
         for(j = 1; j < n-1; j++)
         {
            average = (myVal[i-1][j] + myVal[i][j+1] +
                       myVal[i+1][j] + myVal[i][j-1]) / 4;
            myDelta = Max(myDelta, Abs(average - myVal[i][j]));
            myNew[i][j] = average;
         }
      }
      temp = myNew;   /* prepare for next iteration */
      myNew = myVal;
      myVal = temp;

      pthread_mutex_lock(&update_lock);
      if(myDelta > delta)
      {
         delta = myDelta;   /* update delta */
      }
      pthread_mutex_unlock(&update_lock);

      barrier();  /* wait till every thread has finished its part and updated delta */
      if(id == 0)
      {
         iterates++;    /* count one iteration of the algorithm */
         if (iterates > iterations)
         {
            done = 1;
         }
         else
         {
            done = 0;
            delta = 0.0;  /* reset shared value of delta */
         }
      }
      barrier(); /* wait till thread 0 checks for completion */
   }while(!done);

   return 0;
}/* thread_main() */



void barrier()
{
   pthread_mutex_lock(&barrier_lock);
   count++;
   if(count == t)
   {
      count = 0;
      pthread_cond_broadcast(&all_here);
   }
   else
   {
      pthread_cond_wait(&all_here, &barrier_lock);
   }
   pthread_mutex_unlock(&barrier_lock);
}/* barrier() */
